#include <imsls.h>
#include <stdio.h>
#include <malloc.h>
int main(){
   int i, j, k;                        /* index variables             */
   int n = 50;                         /* population size             */
   int n_generations;                  /* number of generations       */
   int n_nominal = 8;                  /* number of nominal phenotypes*/
   int n_categories[8] = {8, 8, 8, 8,
                          8, 8, 8, 8}; /* nominal category limits     */
   float x1;                           /* temporary storage           */
   float avg;                          /* average fitness             */
   float* genStats;                    /* generation statistics       */
   static float pmxFitness(Imsls_f_individual* individual);
   Imsls_f_chromosome* chromosome;     /* chromosome data structure   */
   Imsls_f_individual* best_individual;/* optimum                     */
   Imsls_f_population* population;     /* population data structure   */
   Imsls_f_population* last_generation;/* last generation             */
   char *cities[8] = {"a", "b", "c", "d",  /* Cities Label a-h        */
                  "e", "f", "g", "h"}; 
   /*******************************************************************/
   /* In this example the user function is thread safe.  Let CNL      */
   /* know it is safe, which allows genetic algorithm to run in       */
   /* parallel, if that capability exists on the user computer.       */
   imsls_omp_options(IMSLS_SET_FUNCTIONS_THREAD_SAFE, 1, 0);
   imsls_random_seed_set(12345);
   chromosome      = imsls_f_ga_chromosome(
               IMSLS_NOMINAL, n_nominal, n_categories, 0);

   population      = imsls_f_ga_random_population(n, chromosome, 
               IMSLS_PMX_CROSSOVER,
               IMSLS_FITNESS_FCN, pmxFitness, 0);
   best_individual = imsls_f_genetic_algorithm(
            pmxFitness, population, 
            IMSLS_PRINT_LEVEL, IMSLS_FINAL,
            IMSLS_PMX_CROSSOVER,
            IMSLS_INVERT_CROSSOVER, 
            IMSLS_CROSSOVER_PROB, 0.8,
            IMSLS_MAX_GENERATIONS, 10,
            IMSLS_GENERATION_STATS, &genStats,
            IMSLS_N_GENERATIONS, &n_generations,
            IMSLS_LAST_GENERATION, &last_generation, 0);

   printf("GENERATION STATISTICS\n");
   printf("Total Number of Generations: %d\n\n", n_generations);
   printf("Generation  Max. Fit. Min. Fit.  Avg. Fit.    CV\n");
   for(i=0; i<=n_generations; i++){
      printf("Gen. %3d: %8.0f %8.0f %12.2f %8.2f\n",
         i, genStats[4*i], genStats[4*i+1], genStats[4*i+2],
         100*genStats[4*i+3]/genStats[4*i+2]);
   }
   printf("\n\n            LAST GENERATION\n");
   printf("*************************************\n");
   printf("Individual  Fitness  Phenotype Values \n");
   avg = last_generation->avgFitness;
   for(i=0; i<last_generation->n; i++){
      x1 = last_generation->fitness[i];
      printf("    %2d     %6.0f    ", i, x1);
      for(j=0; j<last_generation->chromosome->c_length; j++) {
         k = last_generation->individual[i]->nominalPhenotype[j];
         printf("%s ", cities[k]);
      }
      if(x1 == last_generation->maxFitness){
         printf("***\n");
      }else{
         printf("\n");
      }
   }
   printf("Average Fitness: %6.1f\n\n", avg);
   printf("*************************************\n");
   printf("OPTIMUM SOLUTION:\n\nFitness:%4.0f\n\nChromosome:        ", 
      pmxFitness(best_individual));
   for(i=0; i<best_individual->chromosome->c_length; i++) 
      printf("%3d", best_individual->chromosome->allele[i]);
   printf("\n\nPhenotype Values:  ");
   for(i=0; i<n_nominal; i++) {
      k = best_individual->nominalPhenotype[i];
      printf("->%s", cities[k]);
   }
   printf("\n\n");
   printf("freeing best individual\n");
   imsls_f_ga_free_individual(best_individual);
   printf("freeing last generation\n");
   imsls_f_ga_free_population(last_generation);
   printf("freeing chromosome\n");
   free(chromosome);
   printf("freeing population\n");
   imsls_f_ga_free_population(population);
}
/***********************************************************************/
/* FITNESS FUNCTION                                                    */
/********************************************************************* */ 
static float pmxFitness(Imsls_f_individual* individual)
{
   int i=0, k=0, i1, i2; /* Index variables                           */
   int n_nominal = 8;    /* number of nominal phenotypes              */
   float f = 0.0;        /* fitness value                             */
                /* cities:  a    b    c    d    e    f    g   h       */
   float distances[64] = { 0,  17,  27,  73,  61,  57,  51, 23, /* a  */
                      17,   0,  37,  73,  72,  74,  66, 40, /* b  */
                          27,  37,   0,  48,  35,  49,  65, 50, /* c  */
                          73,  73,  48,   0,  47,  82, 113, 95, /* d  */
                          61,  72,  35,  47,   0,  38,  80, 78, /* e  */
                          57,  74,  49,  82,  38,   0,  48, 65, /* f  */
                          51,  66,  65, 113,  80,  48,   0, 40, /* g  */
                          23,  40,  50,  95,  78,  65,  40,  0};/* h  */
   n_nominal = individual->chromosome->n_nominal;
   k = individual->chromosome->nominalIndex+1;
   f = 0.0;
   for(i=k; i<k+n_nominal-1; i++){
      i1 = individual->nominalPhenotype[i-1];
      i2 = individual->nominalPhenotype[i  ];
      f += distances[i1*n_nominal + i2];
   }
   return 516-f;
}
